/*
 * Copyright (c) 2009, 2010, Maciej Smolenski (jezdnia@gmail.com)
 * All rights reserved.
 *
 * This file is part of Oropo project.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 *     * Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *     * Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in the
 *     documentation and/or other materials provided with the distribution.
 *     * Neither the name of the Oropo nor the names of its contributors may
 *     be used to endorse or promote products derived from this software
 *     without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS
 * BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY,
 * OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF
 * THE POSSIBILITY OF SUCH DAMAGE.
 */
#ifndef _OROPO_BASE_QUEUE_HPP_
#define _OROPO_BASE_QUEUE_HPP_

#include <boost/interprocess/sync/named_semaphore.hpp>
#include <boost/interprocess/sync/named_mutex.hpp>
#include <boost/interprocess/sync/scoped_lock.hpp>
#include <boost/interprocess/allocators/allocator.hpp>
#include <boost/interprocess/containers/deque.hpp>
#include <boost/interprocess/managed_shared_memory.hpp>
#include <string>
#include <oropo/base/constructor_helper.hpp>
#include <oropo/base/share.hpp>

namespace oropo
{

namespace bi=boost::interprocess;

template <typename T, typename CT, typename C >
class queue : public share<CT>
{
public:
	typedef T value_type;
protected:
	typedef bi::named_semaphore i_semaphore_t;
	typedef bi::managed_shared_memory::segment_manager segment_manager_t;
	typedef bi::allocator<T, segment_manager_t> allocator_t;
	typedef bi::managed_shared_memory shm_t;

	constructor_helper<i_semaphore_t, CT> m_producer;
	constructor_helper<i_semaphore_t, CT> m_consumer;
	constructor_helper<C *, CT> m_pqueue;

public:
	queue() = delete;
	queue(const char * p_memory_name, const int p_memory_size, const char *p_mutex_name, const char *p_producer_name, const char *p_consumer_name, const int p_count);
	queue(const char * p_memory_name, const char *p_mutex_name, const char *p_producer_name, const char *p_consumer_name);
	queue(const queue &) = delete;
	~queue() = default;
	queue &operator=(const queue &) = delete;

	void put(T &t) throw(bi::interprocess_exception);
	T get() throw(bi::interprocess_exception);
	typename C::size_type size() throw(bi::interprocess_exception);
};

template <typename T, typename CT, typename C>
queue<T, CT, C>::queue(const char * p_memory_name, const int p_memory_size, const char *p_mutex_name, const char *p_producer_name, const char *p_consumer_name, const int p_count) : share<CT>(p_memory_name, p_memory_size, p_mutex_name), m_producer(p_producer_name, p_count), m_consumer(p_consumer_name, 0), m_pqueue(share<CT>::m_memory.m_t, allocator_t(share<CT>::m_memory.m_t.get_segment_manager()))
{
}

template <typename T, typename CT, typename C>
queue<T, CT, C>::queue(const char * p_memory_name, const char *p_mutex_name, const char *p_producer_name, const char *p_consumer_name) : share<CT>(p_memory_name, p_mutex_name), m_producer(p_producer_name), m_consumer(p_consumer_name), m_pqueue(share<CT>::m_memory.m_t)
{
}

template <typename T, typename OCT, typename C>
void queue<T, OCT, C>::put(T &p_t) throw(bi::interprocess_exception)
{
	m_producer.m_t.wait();
	try
	{
		share<OCT>::m_mutex.m_t.lock();
	}
	catch (bi::interprocess_exception &ie)
	{
		m_producer.m_t.post();
		throw ie;
	};
	try
	{
		m_pqueue.m_t->push_back(p_t);
	}
	catch (bi::bad_alloc &ba)
	{
		share<OCT>::m_mutex.m_t.unlock();
		m_producer.m_t.post();
		throw ba;
	};
	share<OCT>::m_mutex.m_t.unlock();
	m_consumer.m_t.post();
}

template <typename T, typename OCT, typename C>
T queue<T, OCT, C>::get() throw(bi::interprocess_exception)
{
	m_consumer.m_t.wait();
	try
	{
		share<OCT>::m_mutex.m_t.lock();
	}
	catch (bi::interprocess_exception &ie)
	{
		m_consumer.m_t.post();
		throw ie;
	};
	try
	{
		T val(m_pqueue.m_t->front());
		m_pqueue.m_t->pop_front();
		share<OCT>::m_mutex.m_t.unlock();
		m_producer.m_t.post();
		return val;
	}
	catch (bi::bad_alloc &ba)
	{
		share<OCT>::m_mutex.m_t.unlock();
		m_consumer.m_t.post();
		throw ba;
	};
}

template <typename T, typename OCT, typename C>
typename C::size_type queue<T, OCT, C>::size() throw(bi::interprocess_exception)
{
	bi::scoped_lock<bi::named_mutex> lock(share<OCT>::m_mutex.m_t);
	return m_pqueue.m_t->size();
}

}

#endif /* #ifndef _OROPO_BASE_QUEUE_HPP_ */
