/*
 * Copyright (c) 2009, 2010, Maciej Smolenski (jezdnia@gmail.com)
 * All rights reserved.
 *
 * This file is part of Oropo project.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 *     * Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *     * Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in the
 *     documentation and/or other materials provided with the distribution.
 *     * Neither the name of the Oropo nor the names of its contributors may
 *     be used to endorse or promote products derived from this software
 *     without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS
 * BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY,
 * OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF
 * THE POSSIBILITY OF SUCH DAMAGE.
 */
#ifndef _OROPO_BASE_SEQUENCE_HPP_
#define _OROPO_BASE_SEQUENCE_HPP_

#include <boost/interprocess/managed_shared_memory.hpp>
#include <boost/interprocess/sync/scoped_lock.hpp>
#include <oropo/base/constructor_helper.hpp>
#include <oropo/base/share.hpp>

namespace oropo
{

namespace bi=boost::interprocess;

template <typename NT>
class sequence_data
{
private:
	NT m_value;
public:
	sequence_data() = default;
	sequence_data(const sequence_data &) = delete;
	~sequence_data() = default;
	sequence_data &operator=(const sequence_data &) = delete;

	void init(bi::managed_shared_memory &m_segment);
	void deinit(bi::managed_shared_memory &m_segment);
	void get_value(NT &p_value) const;
	void set_value(const NT p_value);
};

template <typename NT>
void sequence_data<NT>::init(bi::managed_shared_memory &m_segment)
{
	m_value = 0;
}

template <typename NT>
void sequence_data<NT>::deinit(bi::managed_shared_memory &m_segment)
{
}

template <typename NT>
void sequence_data<NT>::get_value(NT &p_value) const
{
	p_value = m_value;
}

template <typename NT>
void sequence_data<NT>::set_value(const NT p_value)
{
	m_value = p_value;
}

template <typename CT, typename NT = unsigned long>
class sequence_service : public share<CT>
{
protected:
	constructor_initializer_helper<sequence_data<NT> *, CT> m_data;

public:
	sequence_service() = delete;
	sequence_service(const char * p_memory_name, const int p_memory_size, const char *p_mutex_name);
	sequence_service(const char * p_memory_name, const char *p_mutex_name);
	sequence_service(const sequence_service &) = delete;
	~sequence_service() = default;
	sequence_service &operator=(const sequence_service &) = delete;

	void next(NT &value);
};

template <typename CT, typename NT>
void sequence_service<CT, NT>::next(NT &value)
{
	sequence_data<NT> &mdata = *m_data.m_t;
	bi::scoped_lock<bi::named_mutex> mutex(share<CT>::m_mutex.m_t);
	mdata.get_value(value);
	mdata.set_value(value+1);
}

template <typename CT, typename NT>
sequence_service<CT, NT>::sequence_service(const char * p_memory_name, const int p_memory_size, const char *p_mutex_name) : share<CT>(p_memory_name, p_memory_size, p_mutex_name), m_data(share<CT>::m_memory.m_t)
{
}

template <typename CT, typename NT>
sequence_service<CT, NT>::sequence_service(const char * p_memory_name, const char *p_mutex_name) : share<CT>(p_memory_name, p_mutex_name), m_data(share<CT>::m_memory.m_t)
{
}


}

#endif /* #ifndef _OROPO_BASE_SEQUENCE_HPP_ */
