/*
 * Copyright (c) 2009, 2010, Maciej Smolenski (jezdnia@gmail.com)
 * All rights reserved.
 *
 * This file is part of Oropo project.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 *     * Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *     * Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in the
 *     documentation and/or other materials provided with the distribution.
 *     * Neither the name of the Oropo nor the names of its contributors may
 *     be used to endorse or promote products derived from this software
 *     without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS
 * BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY,
 * OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF
 * THE POSSIBILITY OF SUCH DAMAGE.
 */
#ifndef _OROPO_BASE_SIGNAL_HPP_
#define _OROPO_BASE_SIGNAL_HPP_

#include <signal.h>
#include <sys/signalfd.h>
#include <boost/system/system_error.hpp>

namespace oropo
{

namespace bs=boost::system;

class signal_error : public bs::system_error
{
public:
	template <typename ...Args>
	signal_error(Args... args) : bs::system_error(args...) {}
};

struct signal_set
{
	sigset_t set;

	void add_to_the_mask();

	template <typename ...Args>
	void add_to_the_mask(int signum, Args ...args) throw(signal_error);

public:
	template <typename ...Args>
	signal_set(Args ...args) throw(signal_error);
};

template <typename ...Args>
void signal_set::add_to_the_mask(int signum, Args ...args) throw(signal_error)
{
	if (sigaddset(&set, signum)==-1)
		throw signal_error(bs::error_code(errno, bs::get_system_category()), "sigaddset: ");
	add_to_the_mask(args...);
}

template <typename ...Args>
signal_set::signal_set(Args ...args) throw(signal_error)
{
	if (sigemptyset(&set)==-1)
		throw signal_error(bs::error_code(errno, bs::get_system_category()), "sigemptyset: ");

	add_to_the_mask(args...);
}



class signal_fd_handler
{
	signal_set mask;
	int sfd;

public:
	template <typename ...Args>
	signal_fd_handler(Args ...args) throw(signal_error);

	~signal_fd_handler();

	int wait() throw(signal_error);
};

template <typename ...Args>
signal_fd_handler::signal_fd_handler(Args ...args) throw(signal_error) : mask(args...)
{
	if ((sfd = signalfd(-1, &mask.set, 0))==-1)
		throw signal_error(bs::error_code(errno, bs::get_system_category()), "signalfd: ");
}




class signal_blocker
{
	signal_set mask;

public:
	template <typename ...Args>
	signal_blocker(Args ...args) throw(signal_error);

	~signal_blocker();
};

template <typename ...Args>
signal_blocker::signal_blocker(Args ...args) throw(signal_error) : mask(args...)
{
	if (sigprocmask(SIG_BLOCK, &mask.set, NULL)==-1)
		throw signal_error(bs::error_code(errno, bs::get_system_category()), "sigprocmask: ");
}


typedef void (sigaction_info_handler_t)(int, siginfo_t *, void *);
typedef void (sigaction_handler_t)(int);

template <int SIGNUM>
class signal_handler
{
	signal_set mask;
	struct sigaction new_action;
	struct sigaction prev_action;

public:
	template <typename ...Args>
	signal_handler(sigaction_info_handler_t *sh, Args ...args) throw(signal_error);

	template <typename ...Args>
	signal_handler(sigaction_handler_t *sh, Args ...args) throw(signal_error);

	~signal_handler();
};

template <int SIGNUM>
template <typename ...Args>
signal_handler<SIGNUM>::signal_handler(sigaction_info_handler_t sh, Args ...args) throw(signal_error) : mask(args...)
{
	new_action.sa_sigaction = sh;
	new_action.sa_mask = mask.set;
	new_action.sa_flags = SA_SIGINFO;

	if (sigaction(SIGNUM, &new_action, &prev_action)==-1)
		throw signal_error(bs::error_code(errno, bs::get_system_category()), "sigaction failed: ");
}

template <int SIGNUM>
template <typename ...Args>
signal_handler<SIGNUM>::signal_handler(sigaction_handler_t sh, Args ...args) throw(signal_error) : mask(args...)
{
	new_action.sa_handler = sh;
	new_action.sa_mask = mask.set;
	new_action.sa_flags = 0;

	if (sigaction(SIGNUM, &new_action, &prev_action)==-1)
		throw signal_error(bs::error_code(errno, bs::get_system_category()), "sigaction failed: ");
}

template <int SIGNUM>
signal_handler<SIGNUM>::~signal_handler()
{
	sigaction(SIGNUM, &prev_action, NULL);
}


}

#endif /* #ifndef _OROPO_BASE_SIGNAL_HPP_ */
