/*
 * Copyright (c) 2009, 2010, Maciej Smolenski (jezdnia@gmail.com)
 * All rights reserved.
 *
 * This file is part of Oropo project.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 *     * Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *     * Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in the
 *     documentation and/or other materials provided with the distribution.
 *     * Neither the name of the Oropo nor the names of its contributors may
 *     be used to endorse or promote products derived from this software
 *     without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS
 * BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY,
 * OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF
 * THE POSSIBILITY OF SUCH DAMAGE.
 */
#ifndef _OROPO_MANAGER_CLIENT_HPP_
#define _OROPO_MANAGER_CLIENT_HPP_

#include <iterator>
#include <vector>
#include <string.h>
#include <boost/scoped_array.hpp>
#include <protocol/oropo.h>
#include <oropo/rpc/client.hpp>
#include <oropo/manager/parameter.hpp>
#include <oropo/rpc/error.hpp>
#include <oropo/base/signal.hpp>

namespace oropo
{
namespace manager
{

namespace oro = oropo;
namespace orom = oro::manager;
namespace oror = oropo::rpc;

template <typename CT>
class client : public oror::client<CT>
{
public:
	client(const std::string p_host, const long unsigned p_program, const long unsigned p_version) throw(oror::error);
	template <typename II, typename OI>
	void call(const long unsigned p_procedure, const long unsigned p_timeout, const II p_in_parameters_begin, const II p_in_parameters_end, OI p_out_parameters_iter) throw(oror::error, orom::parameter_error);
};

template <typename CT>
client<CT>::client(const std::string p_host, const long unsigned p_program, const long unsigned p_version) throw(oror::error) : oror::client<CT>(p_host, p_program, p_version)
{
}

template <typename CT>
template <typename II, typename OI>
void client<CT>::call(const long unsigned p_procedure, const long unsigned p_timeout, const II p_in_parameters_begin, const II p_in_parameters_end, OI p_out_parameters_iter) throw(oror::error, orom::parameter_error)
{
	oro::signal_handler<SIGPIPE> si(SIG_IGN);
	orom::protocol_handler handler;

	int count = 0;
	for (II iter = p_in_parameters_begin; iter != p_in_parameters_end; iter++)
		count++;

	request_t request;
	boost::scoped_array<parameter_t> sp_array(new parameter_t[count]);
	if (sp_array.get() == NULL) throw std::bad_alloc();
	request.parameters_t_val = sp_array.get();
	request.parameters_t_len = count;

	handler.memory_to_protocol(p_in_parameters_begin, p_in_parameters_end, request.parameters_t_val);

	response_t response;
	memset((char *) &response, 0 , sizeof(response));

	struct timeval timeout;
	timeout.tv_sec = p_timeout;
	timeout.tv_usec = 0;

	enum clnt_stat call_status;
	try
	{
		call_status = clnt_call(oror::client<CT>::m_client, p_procedure, (xdrproc_t) xdr_request_t, (caddr_t) &request, (xdrproc_t) xdr_response_t, (caddr_t) &response, timeout);
	}
	catch (oror::error &re)
	{
		handler.free_pro_memory(request.parameters_t_val, request.parameters_t_val+request.parameters_t_len);
		throw re;
	}

	handler.free_pro_memory(request.parameters_t_val, request.parameters_t_val+request.parameters_t_len);

	if (call_status != RPC_SUCCESS)
		throw oror::error(oror::make_error_code(call_status), "clnt_call");

	typedef std::vector<orom::mem_parameter> rv_t;
	rv_t result_params;
	try
	{
		handler.protocol_to_memory(response.parameters_t_val, response.parameters_t_val+response.parameters_t_len, std::back_insert_iterator<rv_t>(result_params));
	}
	catch (orom::parameter_error &pe)
	{
		if (clnt_freeres(oror::client<CT>::m_client, (xdrproc_t) xdr_response_t, (caddr_t) &response) == FALSE)
			throw oror::error(oror::make_error_code(RPC_FAILED), "clnt_freeres");
		throw pe;
	}

	if (clnt_freeres(oror::client<CT>::m_client, (xdrproc_t) xdr_response_t, (caddr_t) &response) == FALSE)
	{
		handler.free_resources(result_params.begin(), result_params.end());
		throw oror::error(oror::make_error_code(RPC_FAILED), "clnt_freeres");
	}

	std::copy(result_params.begin(), result_params.end(), p_out_parameters_iter);
}

}
}

#endif /* #ifndef _OROPO_MANAGER_CLIENT_HPP_ */
