/*
 * Copyright (c) 2009, 2010, Maciej Smolenski (jezdnia@gmail.com)
 * All rights reserved.
 *
 * This file is part of Oropo project.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 *     * Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *     * Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in the
 *     documentation and/or other materials provided with the distribution.
 *     * Neither the name of the Oropo nor the names of its contributors may
 *     be used to endorse or promote products derived from this software
 *     without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS
 * BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY,
 * OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF
 * THE POSSIBILITY OF SUCH DAMAGE.
 */
#ifndef _PARAMETER_HPP_
#define _PARAMETER_HPP_


#include <protocol/oropo.h>
#include <iostream>
#include <string>
#include <vector>
#include <oropo/base/common.hpp>
#include <oropo/base/datafile_converter.hpp>
#include <oropo/manager/parameter_error.hpp>


namespace oropo
{
namespace manager
{

namespace bi = boost::interprocess;

enum class param_t
{
	PARAM_NONE, PARAM_STRING, PARAM_INT, PARAM_FLOAT, PARAM_DATA, PARAM_PATH
};

std::ostream &operator<<(std::ostream &os, const param_t &p_type);
std::istream &operator>>(std::istream &is, param_t &p_type);

class shm_parameter;
class mem_parameter;
class protocol_handler;

class parameter
{
protected:
	param_t m_type;
public:
	parameter(const param_t p_type = param_t::PARAM_NONE);
	parameter(const parameter &) = default;
	~parameter() = default;
	parameter &operator=(const parameter &) = default;
	param_t get_type() const;
	void set_type(param_t const p_type);

	template <typename Archive>
	friend void serialize(Archive &ar, parameter &p_param, const unsigned int version);
};

class shm_parameter : public parameter
{
protected:
	shm_string m_string;
	int m_int;
	float m_float;

public:
	shm_parameter() = delete;
	shm_parameter(const shm_char_allocator &p_ca, const std::string &p_string, const param_t p_type = param_t::PARAM_STRING);
	shm_parameter(const shm_char_allocator &p_ca, const shm_string &p_string, const param_t p_type = param_t::PARAM_STRING);
	shm_parameter(const shm_char_allocator &p_ca, const int &p_int);
	shm_parameter(const shm_char_allocator &p_ca, const float &p_float);
	shm_parameter(const shm_char_allocator &p_ca, const parameter_t &p_pt) throw(parameter_error);
	shm_parameter(const shm_char_allocator &p_ca, const mem_parameter &p_rmp);
	shm_parameter(const shm_parameter &) = default;
	~shm_parameter() = default;
	shm_parameter &operator=(const shm_parameter &) = default;

public:
	shm_string get_string();
	int get_int();
	float get_float();
	shm_string get_path();

	friend std::ostream &operator<<(std::ostream &co, const shm_parameter &p_rpc);
	friend class mem_parameter;
	friend class protocol_handler;
};

class mem_parameter : public parameter
{
	std::string m_string;
	int m_int;
	float m_float;

public:
	mem_parameter() = default;
	mem_parameter(const mem_parameter &) = default;
	~mem_parameter() = default;
	mem_parameter &operator=(const mem_parameter &) = default;

	mem_parameter(const std::string &p_string, const param_t p_type = param_t::PARAM_STRING);
	mem_parameter(const int &p_int);
	mem_parameter(const float &p_float);
	mem_parameter(const parameter_t &p_pt) throw(parameter_error);
	mem_parameter(const shm_parameter &p_rsp);

	std::string get_string();
	int get_int();
	float get_float();
	std::string get_path();

	template <typename Archive>
	friend void serialize(Archive &ar, mem_parameter &p_mem_param, const unsigned int version);
	friend std::ostream &operator<<(std::ostream &co, const mem_parameter &p_rpc);
	friend std::istream &operator>>(std::istream &ci, mem_parameter &p_parameter);
	friend class shm_parameter;
	friend class protocol_handler;
};

class protocol_handler
{
public:
	protocol_handler() = default;
	protocol_handler(const protocol_handler &) = delete;
	~protocol_handler() = default;
	protocol_handler &operator=(const protocol_handler &) = delete;

	mem_parameter protocol_to_memory(const struct parameter_t &) throw(parameter_error);
	struct parameter_t memory_to_protocol(const mem_parameter &) throw(parameter_error, std::bad_alloc);
	void free_pro_memory(struct parameter_t &) throw(parameter_error);
	void free_resources(const mem_parameter &) throw(parameter_error);
	template <typename I, typename BI>
	void protocol_to_memory(const I it_begin, const I it_end, BI p_bi_i) throw(parameter_error);
	template <typename I, typename BI>
	void memory_to_protocol(const I it_begin, const I it_end, BI p_bi_i) throw(parameter_error, std::bad_alloc);
	template <typename I>
	void free_pro_memory(const I it_begin, const I it_end) throw(parameter_error);
	template <typename I>
	void free_resources(const I it_begin, const I it_end) throw(parameter_error);
};

template <typename I, typename BI>
void protocol_handler::protocol_to_memory(const I it_begin, const I it_end, BI p_bi_i) throw(parameter_error)
{
	typedef std::vector<mem_parameter> vmp_t;
	vmp_t tmp;
	try
	{
		for (I iter = it_begin; iter != it_end; iter++)
			tmp.push_back(protocol_to_memory(*iter));
	}
	catch (...)
	{
		free_resources(tmp.begin(), tmp.end());
		throw parameter_error(make_error_code(parameter_error_enum_t::p2m_conversion), "protocol_to_memory");
	}
	for (vmp_t::iterator iter = tmp.begin(); iter != tmp.end(); iter++)
		*p_bi_i++ = *iter;
}

template <typename I, typename BI>
void protocol_handler::memory_to_protocol(const I it_begin, const I it_end, BI p_bi_i) throw(parameter_error, std::bad_alloc)
{
	typedef std::vector<parameter_t> vmp_t;
	vmp_t tmp;
	try
	{
		for (I iter = it_begin; iter != it_end; iter++)
			tmp.push_back(memory_to_protocol(*iter));
	}
	catch (...)
	{
		free_pro_memory(tmp.begin(), tmp.end());
		throw parameter_error(make_error_code(parameter_error_enum_t::m2p_conversion), "memory_to_protocol");
	}
	for (vmp_t::iterator iter = tmp.begin(); iter != tmp.end(); iter++)
		*p_bi_i++ = *iter;
}

template <typename I>
void protocol_handler::free_pro_memory(const I it_begin, const I it_end) throw(parameter_error)
{
	for (I iter = it_begin; iter != it_end; iter++)
		free_pro_memory(*iter);
}

template <typename I>
void protocol_handler::free_resources(const I it_begin, const I it_end) throw(parameter_error)
{
	for (I iter = it_begin; iter != it_end; iter++)
		free_resources(*iter);
}

}
}

#endif /* #ifndef _PARAMETER_HPP_ */
