/*
 * Copyright (c) 2009, 2010, Maciej Smolenski (jezdnia@gmail.com)
 * All rights reserved.
 *
 * This file is part of Oropo project.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 *     * Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *     * Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in the
 *     documentation and/or other materials provided with the distribution.
 *     * Neither the name of the Oropo nor the names of its contributors may
 *     be used to endorse or promote products derived from this software
 *     without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS
 * BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY,
 * OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF
 * THE POSSIBILITY OF SUCH DAMAGE.
 */
#ifndef _OROPO_MANAGER_REQUEST_HPP_
#define _OROPO_MANAGER_REQUEST_HPP_

#include <oropo/manager/parameter.hpp>
#include <oropo/manager/task.hpp>
#include <boost/interprocess/allocators/allocator.hpp>
#include <boost/interprocess/containers/vector.hpp>
#include <iostream>
#include <vector>

namespace oropo
{
namespace manager
{

namespace bi=boost::interprocess;

typedef bi::allocator<shm_parameter, bi::managed_shared_memory::segment_manager> parameter_allocator;
typedef bi::vector<shm_parameter, parameter_allocator> shm_parameters_vector;

template <typename T>
class request : public task<T>
{
private:
	unsigned long m_id;
	unsigned long m_program;
	unsigned long m_version;
	unsigned long m_procedure;
	unsigned long m_timeout;
public:
	request() = default;
	request(const typename task<T>::params_t &p_parameters, const unsigned long p_id, const unsigned long p_program, const unsigned long p_version, const unsigned long p_procedure, const unsigned long p_timeout);
	request(const request &) = default;
	~request() = default;
	request &operator=(const request &) = default;

	unsigned long get_id() const;
	unsigned long get_program() const;
	unsigned long get_version() const;
	unsigned long get_procedure() const;
	unsigned long get_timeout() const;

	template <typename Archive, typename E>
	friend void serialize(Archive &ar, request<E> &p_param, const unsigned int version);
	template <typename E>
	friend std::ostream &operator<<(std::ostream &os, const request<E> &p_request);
};

template <typename T>
request<T>::request(const typename task<T>::params_t &p_parameters, const unsigned long p_id, const unsigned long p_program, const unsigned long p_version, const unsigned long p_procedure, const unsigned long p_timeout) : task<T>(p_parameters), m_id(p_id), m_program(p_program), m_version(p_version), m_procedure(p_procedure), m_timeout(p_timeout) {}

template <typename T>
unsigned long request<T>::get_id() const
{
	return m_id;
}

template <typename T>
unsigned long request<T>::get_program() const
{
	return m_program;
}

template <typename T>
unsigned long request<T>::get_version() const
{
	return m_version;
}

template <typename T>
unsigned long request<T>::get_procedure() const
{
	return m_procedure;
}

template <typename T>
unsigned long request<T>::get_timeout() const
{
	return m_timeout;
}

class mem_request;
class shm_request;

class mem_request : public request<std::vector<mem_parameter> >
{
public:
	mem_request() = default;
	mem_request(const params_t &p_parameters, const unsigned long p_id, const unsigned long p_program, const unsigned long p_version, const unsigned long p_procedure, const unsigned long p_timeout);
	mem_request(const shm_request &p_st);
	mem_request(const mem_request &) = default;
	~mem_request() = default;
	mem_request &operator=(const mem_request &) = default;

	template <typename Archive>
	friend void serialize(Archive &ar, mem_request &p_param, const unsigned int version);
	friend std::ostream &operator<<(std::ostream &os, const mem_request &p_request);
	friend class shm_request;
};

class shm_request : public request<shm_parameters_vector>
{
public:
	shm_request() = delete;
	shm_request(const params_t &p_parameters, const unsigned long p_id, const unsigned long p_program, const unsigned long p_version, const unsigned long p_procedure, const unsigned long p_timeout);
	shm_request(bi::managed_shared_memory::segment_manager &p_sm, const mem_request &p_mt);
	shm_request(const shm_request &) = default;
	~shm_request() = default;
	shm_request &operator=(const shm_request &) = default;

	friend std::ostream &operator<<(std::ostream &os, const shm_request &p_request);
	friend class mem_request;
};

template <typename E>
std::ostream &operator<<(std::ostream &os, const request<E> &p_request)
{
	os << static_cast<const task<E> &>(p_request);
	os << "id: " << p_request.m_id;
	os << " program: " << p_request.m_program;
	os << " version: " << p_request.m_version;
	os << " procedure: " << p_request.m_procedure;
	os << " timeout: " << p_request.m_timeout;
	os << " ";
	return os;
}

}
}

#endif /* #ifndef _OROPO_MANAGER_REQUEST_HPP_ */
