/*
 * Copyright (c) 2009, 2010, Maciej Smolenski (jezdnia@gmail.com)
 * All rights reserved.
 *
 * This file is part of Oropo project.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 *     * Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *     * Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in the
 *     documentation and/or other materials provided with the distribution.
 *     * Neither the name of the Oropo nor the names of its contributors may
 *     be used to endorse or promote products derived from this software
 *     without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS
 * BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY,
 * OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF
 * THE POSSIBILITY OF SUCH DAMAGE.
 */
#ifndef _OROPO_MANAGER_RESPONSE_HPP_
#define _OROPO_MANAGER_RESPONSE_HPP_

#include <oropo/manager/parameter.hpp>
#include <oropo/manager/task.hpp>
#include <boost/interprocess/allocators/allocator.hpp>
#include <boost/interprocess/containers/vector.hpp>
#include <iostream>
#include <vector>

namespace oropo
{
namespace manager
{

namespace bi=boost::interprocess;

enum result_t { result_ok, result_failure, result_unavailable };

typedef bi::allocator<shm_parameter, bi::managed_shared_memory::segment_manager> parameter_allocator;
typedef bi::vector<shm_parameter, parameter_allocator> shm_parameters_vector;


template <typename T>
class response : public task<T>
{
private:
	unsigned long m_id;
	result_t m_result;
public:
	response() = default;
	response(const typename task<T>::params_t &p_parameters, const unsigned long p_id, const result_t p_result);
	response(const response &) = default;
	~response() = default;
	response &operator=(const response &) = default;

	unsigned long get_id() const;
	result_t get_result() const;

	template <typename Archive, typename E>
	friend void serialize(Archive &ar, response<E> &p_response, const unsigned int version);
	template <typename E>
	friend std::ostream &operator<<(std::ostream &os, const response<E> &p_response);
};

template <typename T>
response<T>::response(const typename task<T>::params_t &p_parameters, const unsigned long p_id, const result_t p_result) : task<T>(p_parameters), m_id(p_id), m_result(p_result)
{
}

template <typename T>
unsigned long response<T>::get_id() const
{
	return m_id;
}

template <typename T>
result_t response<T>::get_result() const
{
	return m_result;
}

class mem_response;
class shm_response;

class mem_response : public response<std::vector<mem_parameter> >
{
public:
	mem_response() = default;
	mem_response(const params_t &p_parameters, const unsigned long p_id, const result_t p_result);
	mem_response(const shm_response &p_sr);
	mem_response(const mem_response &) = default;
	~mem_response() = default;
	mem_response &operator=(const mem_response &) = default;

	template <typename Archive>
	friend void serialize(Archive &ar, mem_response &p_mem_response, const unsigned int version);
	friend std::ostream &operator<<(std::ostream &os, const mem_response &p_response);
	friend class shm_response;
};

class shm_response : public response<shm_parameters_vector>
{
public:
	shm_response() = delete;
	shm_response(const params_t &p_parameters, const unsigned long p_id, const result_t p_result);
	shm_response(bi::managed_shared_memory::segment_manager &p_sm, const mem_response &p_mr);
	shm_response(const shm_response &) = default;
	~shm_response() = default;
	shm_response &operator=(const shm_response &) = default;

	friend std::ostream &operator<<(std::ostream &os, const shm_response &p_response);
	friend class mem_response;
};

template <typename E>
std::ostream &operator<<(std::ostream &os, const response<E> &p_response)
{
	os << static_cast<const task<E> &>(p_response);
	os << "id: " << p_response.m_id;
	os << " result: " << p_response.m_result;
	os << " ";
	return os;
}

}
}


#endif /* #ifndef _OROPO_MANAGER_RESPONSE_HPP_ */
