/*
 * Copyright (c) 2009, 2010, Maciej Smolenski (jezdnia@gmail.com)
 * All rights reserved.
 *
 * This file is part of Oropo project.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 *     * Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *     * Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in the
 *     documentation and/or other materials provided with the distribution.
 *     * Neither the name of the Oropo nor the names of its contributors may
 *     be used to endorse or promote products derived from this software
 *     without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS
 * BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY,
 * OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF
 * THE POSSIBILITY OF SUCH DAMAGE.
 */
#ifndef _OROPO_MANAGER_SERVICE_HPP_
#define _OROPO_MANAGER_SERVICE_HPP_

#include <string>
#include <vector>
#include <rpc/pmap_clnt.h>
#include <boost/scoped_array.hpp>
#include <protocol/oropo.h>
#include <oropo/manager/parameter.hpp>
#include <oropo/rpc/error.hpp>
#include <oropo/base/signal.hpp>

namespace oropo
{
namespace manager
{

namespace oro=oropo;
namespace oror=oro::rpc;
namespace orom=oro::manager;

class service_base
{
	unsigned long m_procedure;
protected:
	service_base() = delete;
	service_base(unsigned long p_procedure);
	service_base(const service_base &) = delete;
	~service_base() = default;
	service_base &operator=(const service_base &) = delete;
public:
	unsigned long get_procedure();

	/*
		template <typename II, typename OI>
		void operator()(const II in_params_begin, const II in_params_end, OI out_params_iter);
		template <typename II>
		void free_resources(const II out_params_begin, const II out_params_end);
	*/

};

typedef std::vector<orom::mem_parameter> vp_t;


template <typename T>
static void protocol_service(struct svc_req *rqstp, register SVCXPRT *transport) throw(oror::error)
{
	oro::signal_handler<SIGPIPE> si(SIG_IGN);
	unsigned long procedure = rqstp->rq_proc;
	T &service(T::Instance());

	if (procedure == NULLPROC)
	{
		if (!svc_sendreply(transport, (xdrproc_t) xdr_void, (char *)NULL))
			throw oror::error(oror::make_error_code(RPC_FAILED), "svc_sendreply");
	}
	else if (procedure == service.get_procedure())
	{
		request_t request;
		memset((char *)&request, 0, sizeof(request));

		if (!svc_getargs(transport, (xdrproc_t) xdr_request_t, (caddr_t) &request))
		{
			svcerr_decode(transport);
			throw oror::error(oror::make_error_code(RPC_FAILED), "svc_getargs");
		}

		orom::protocol_handler handler;

		vp_t vrp_req;
		std::back_insert_iterator<vp_t> vrp_req_iter(vrp_req);
		handler.protocol_to_memory(request.parameters_t_val, request.parameters_t_val+request.parameters_t_len, vrp_req_iter);

		if (!svc_freeargs(transport, (xdrproc_t) xdr_request_t, (caddr_t) &request))
		{
			handler.free_resources(vrp_req.begin(), vrp_req.end());
			throw oror::error(oror::make_error_code(RPC_FAILED), "svc_freeargs");
		}

		vp_t vrp_res;
		std::back_insert_iterator<vp_t> vrp_res_iter(vrp_res);

		service(vrp_req.begin(), vrp_req.end(), vrp_res_iter);
		handler.free_resources(vrp_req.begin(), vrp_req.end());

		response_t result = {0, NULL};

		int param_count = vrp_res.size();
		boost::scoped_array<parameter_t> sp_array(new parameter_t[param_count]);
		if (sp_array.get()==NULL) throw std::bad_alloc();
		result.parameters_t_val = sp_array.get();
		result.parameters_t_len = param_count;

		handler.memory_to_protocol(vrp_res.begin(), vrp_res.end(), result.parameters_t_val);

		if (!svc_sendreply(transport, (xdrproc_t) xdr_response_t, (caddr_t) &result))
		{
			service.free_resources(vrp_res.begin(), vrp_res.end());
			handler.free_pro_memory(result.parameters_t_val, result.parameters_t_val+result.parameters_t_len);
			throw oror::error(oror::make_error_code(RPC_FAILED), "svc_sendreply");
		}

		service.free_resources(vrp_res.begin(), vrp_res.end());
		handler.free_pro_memory(result.parameters_t_val, result.parameters_t_val+result.parameters_t_len);
	}
	else
		svcerr_noproc(transport);
}

template <typename T>
static void protocol_service_catcher(struct svc_req *rqstp, register SVCXPRT *transport)
{
	try
	{
		protocol_service<T>(rqstp, transport);
	}
	catch (oror::error &e)
	{
		std::cerr << e.what() << std::endl;
	}
}

template <typename CT, typename T>
static void service_protocol_run(const unsigned long p_program, const unsigned long p_version) throw(oror::error)
{
	register SVCXPRT *transport;

	pmap_unset(p_program, p_version);

	transport = CT::create_service(RPC_ANYSOCK);
	if (transport == NULL)
		throw oror::error(oror::make_error_code(RPC_FAILED), "create_service");
	if (!CT::register_service(transport, p_program, p_version, protocol_service_catcher<T>))
		throw oror::error(oror::make_error_code(RPC_FAILED), "svc_register");

	svc_run();
	throw oror::error(oror::make_error_code(RPC_FAILED), "svc_run");
}

}
}

#endif /* #ifndef _OROPO_MANAGER_SERVICE_HPP_ */
