/*
 * Copyright (c) 2009, 2010, Maciej Smolenski (jezdnia@gmail.com)
 * All rights reserved.
 *
 * This file is part of Oropo project.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 *     * Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *     * Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in the
 *     documentation and/or other materials provided with the distribution.
 *     * Neither the name of the Oropo nor the names of its contributors may
 *     be used to endorse or promote products derived from this software
 *     without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS
 * BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY,
 * OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF
 * THE POSSIBILITY OF SUCH DAMAGE.
 */
#ifndef _OROPO_MONITOR_CONFIG_HPP_
#define _OROPO_MONITOR_CONFIG_HPP_

#include <oropo/monitor/config_data.hpp>
#include <oropo/base/constructor_helper.hpp>
#include <oropo/base/share.hpp>
#include <boost/interprocess/managed_shared_memory.hpp>
#include <boost/interprocess/sync/scoped_lock.hpp>

namespace oropo
{

namespace monitor
{

namespace bi=boost::interprocess;

template <typename CT>
class config_service : public share<CT>
{
	constructor_initializer_helper<config *, CT> m_config;
public:
	config_service() = delete;
	config_service(const char * p_memory_name, const int p_memory_size, const char *p_mutex_name);
	config_service(const char * p_memory_name, const char *p_mutex_name);
	config_service(const config_service &) = delete;
	~config_service() = default;
	config_service &operator=(const config_service &) = delete;

	void get_hosts(config_stdset &so);
	void set_hosts(config_stdset &si);
	void add_host(host_entry_string &he);
	void del_host(host_entry_string &he);
	void get_delay(long &p_delay);
	void set_delay(long &p_delay);
	void get_ping_timeout(long &p_ping_timeout);
	void set_ping_timeout(long &p_ping_timeout);
	void get_portmap_timeout(long &p_portmap_timeout);
	void set_portmap_timeout(long &p_portmap_timeout);
	void get_pgroup(int &p_portmap_timeout);
	void set_pgroup(int &p_portmap_timeout);
private:
	template <typename FT>
	void get_field(FT config::* field, FT &value);
	template <typename FT>
	void set_field(FT config::* field, FT &value);
};

template <typename CT>
config_service<CT>::config_service(const char *p_memory_name, const int p_memory_size, const char *p_mutex_name) : share<CT>(p_memory_name, p_memory_size, p_mutex_name), m_config(share<CT>::m_memory.m_t)
{
}

template <typename CT>
config_service<CT>::config_service(const char *p_memory_name, const char *p_mutex_name) : share<CT>(p_memory_name, p_mutex_name), m_config(share<CT>::m_memory.m_t)
{
}

template <typename CT>
void config_service<CT>::get_hosts(config_stdset &so)
{
	bi::managed_shared_memory &segment = share<CT>::m_memory.m_t;
	config &mconfig = *m_config.m_t;
	bi::scoped_lock<bi::named_mutex> mutex(share<CT>::m_mutex.m_t);
	bi_set_to_std_set(segment, *mconfig.m_hosts, so);
}

template <typename CT>
void config_service<CT>::set_hosts(config_stdset &si)
{
	bi::managed_shared_memory &segment = share<CT>::m_memory.m_t;
	config &mconfig = *m_config.m_t;
	bi::scoped_lock<bi::named_mutex> mutex(share<CT>::m_mutex.m_t);
	std_set_to_bi_set(segment, si, *mconfig.m_hosts);
}

template <typename CT>
void config_service<CT>::add_host(host_entry_string &he)
{
	bi::managed_shared_memory &segment = share<CT>::m_memory.m_t;
	config &mconfig = *m_config.m_t;
	bi::scoped_lock<bi::named_mutex> mutex(share<CT>::m_mutex.m_t);
	shm_char_allocator ca(segment.get_segment_manager());
	host_entry_shm_string hes(he, ca);
	mconfig.m_hosts->insert(hes);
}

template <typename CT>
void config_service<CT>::del_host(host_entry_string &he)
{
	bi::managed_shared_memory &segment = share<CT>::m_memory.m_t;
	config &mconfig = *m_config.m_t;
	bi::scoped_lock<bi::named_mutex> mutex(share<CT>::m_mutex.m_t);
	shm_char_allocator ca(segment.get_segment_manager());
	host_entry_shm_string hes(he, ca);
	mconfig.m_hosts->erase(hes);
}




template <typename CT>
template <typename FT>
void config_service<CT>::get_field(FT config::* field, FT &value)
{
	bi::managed_shared_memory &segment = share<CT>::m_memory.m_t;
	config &mconfig = *m_config.m_t;
	bi::scoped_lock<bi::named_mutex> mutex(share<CT>::m_mutex.m_t);
	value = mconfig.*field;
}


template <typename CT>
template <typename FT>
void config_service<CT>::set_field(FT config::* field, FT &value)
{
	bi::managed_shared_memory &segment = share<CT>::m_memory.m_t;
	config &mconfig = *m_config.m_t;
	bi::scoped_lock<bi::named_mutex> mutex(share<CT>::m_mutex.m_t);
	mconfig.*field = value;
}


template <typename CT>
void config_service<CT>::get_delay(long &p_delay)
{
	get_field<long>(&config::m_delay, p_delay);
}


template <typename CT>
void config_service<CT>::set_delay(long &p_delay)
{
	set_field<long>(&config::m_delay, p_delay);
}

template <typename CT>
void config_service<CT>::get_ping_timeout(long &p_ping_timeout)
{
	get_field<long>(&config::m_ping_timeout, p_ping_timeout);
}


template <typename CT>
void config_service<CT>::set_ping_timeout(long &p_ping_timeout)
{
	set_field<long>(&config::m_ping_timeout, p_ping_timeout);
}

template <typename CT>
void config_service<CT>::get_portmap_timeout(long &p_portmap_timeout)
{
	get_field<long>(&config::m_portmap_timeout, p_portmap_timeout);
}


template <typename CT>
void config_service<CT>::set_portmap_timeout(long &p_portmap_timeout)
{
	set_field<long>(&config::m_portmap_timeout, p_portmap_timeout);
}

template <typename CT>
void config_service<CT>::get_pgroup(int &p_pgroup)
{
	get_field<int>(&config::m_pgroup, p_pgroup);
}


template <typename CT>
void config_service<CT>::set_pgroup(int &p_pgroup)
{
	set_field<int>(&config::m_pgroup, p_pgroup);
}

}

}

#endif

