/*
 * Copyright (c) 2009, 2010, Maciej Smolenski (jezdnia@gmail.com)
 * All rights reserved.
 *
 * This file is part of Oropo project.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 *     * Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *     * Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in the
 *     documentation and/or other materials provided with the distribution.
 *     * Neither the name of the Oropo nor the names of its contributors may
 *     be used to endorse or promote products derived from this software
 *     without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS
 * BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY,
 * OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF
 * THE POSSIBILITY OF SUCH DAMAGE.
 */
#ifndef _OROPO_MONITOR_CONFIG_DATA_HPP_
#define _OROPO_MONITOR_CONFIG_DATA_HPP_

#include <boost/interprocess/managed_shared_memory.hpp>
#include <boost/interprocess/containers/string.hpp>
#include <boost/interprocess/containers/set.hpp>
#include <boost/interprocess/sync/interprocess_mutex.hpp>
#include <boost/interprocess/offset_ptr.hpp>

#include <string>
#include <set>

#include <oropo/base/common.hpp>

namespace oropo
{

namespace monitor
{

namespace oro=oropo;

template <typename AT>
struct host_entry
{
	AT address;
	long port;
	protocol_t protocol;

	host_entry() = delete;
	host_entry(const AT p_address, const long p_port, const protocol_t p_protocol);
	host_entry(const host_entry<AT> &he) = default;
	~host_entry() = default;
	host_entry &operator=(const host_entry &) = default;

	bool operator<(const host_entry<AT> &he) const;
};

template <typename AT>
host_entry<AT>::host_entry(const AT p_address, const long p_port, const protocol_t p_protocol) : address(p_address), port(p_port), protocol(p_protocol) {}

template <typename AT>
bool host_entry<AT>::operator<(const host_entry<AT> &he) const
{
	if (address.compare(he.address)<0) return true;
	else if (port<he.port) return true;
	else if (protocol<he.protocol) return true;
	else return false;
}

struct host_entry_string;
struct host_entry_shm_string;

struct host_entry_string : public host_entry<std::string>
{
	host_entry_string() = delete;
	host_entry_string(const std::string p_address, const long p_port, const protocol_t p_protocol);
	host_entry_string(const host_entry_shm_string &hes);
	host_entry_string(const host_entry_string &) = default;
	~host_entry_string() = default;
	host_entry_string &operator=(const host_entry_string &) = default;
};

struct host_entry_shm_string : public host_entry<shm_string>
{
	host_entry_shm_string() = delete;
	host_entry_shm_string(const shm_string p_address, const long p_port, const protocol_t p_protocol);
	host_entry_shm_string(const host_entry_string &hes, oro::shm_char_allocator &al);
	host_entry_shm_string(const host_entry_shm_string &) = default;
	~host_entry_shm_string() = default;
	host_entry_shm_string &operator=(const host_entry_shm_string &) = default;
};

typedef std::set<host_entry_string> config_stdset;
typedef bi::managed_shared_memory::allocator<host_entry_shm_string>::type shm_host_entry_shm_string_allocator;
typedef bi::set<host_entry_shm_string, std::less<host_entry_shm_string>, shm_host_entry_shm_string_allocator> config_biset;

void std_set_to_bi_set(bi::managed_shared_memory &segment, const config_stdset &si, config_biset &so);
void bi_set_to_std_set(bi::managed_shared_memory &segment, const config_biset &si, config_stdset &so);


struct config
{
	long m_delay;
	long m_ping_timeout;
	long m_portmap_timeout;
	int m_pgroup;
	bi::offset_ptr<config_biset> m_hosts;

	config() = default;
	config(const config &) = delete;
	~config() = default;
	config &operator=(const config &) = delete;

	void init(bi::managed_shared_memory &m_segment, const long p_delay = 60, const long p_ping_timeout = 2, const long p_portmap_timeout = 2, const int p_pgroup = 0);
	void deinit(bi::managed_shared_memory &m_segment);
};

std::ostream &operator<<(std::ostream &os, const host_entry_string &r);

}
}

#endif
