/*
 * Copyright (c) 2009, 2010, Maciej Smolenski (jezdnia@gmail.com)
 * All rights reserved.
 *
 * This file is part of Oropo project.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 *     * Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *     * Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in the
 *     documentation and/or other materials provided with the distribution.
 *     * Neither the name of the Oropo nor the names of its contributors may
 *     be used to endorse or promote products derived from this software
 *     without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS
 * BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY,
 * OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF
 * THE POSSIBILITY OF SUCH DAMAGE.
 */
#ifndef _OROPO_MONITOR_MONITOR_DATA_HPP_
#define _OROPO_MONITOR_MONITOR_DATA_HPP_

#include <boost/interprocess/managed_shared_memory.hpp>
#include <boost/interprocess/containers/set.hpp>
#include <boost/interprocess/offset_ptr.hpp>

#include <string>
#include <set>

#include <oropo/base/common.hpp>

namespace oropo
{

namespace monitor
{

namespace bi = boost::interprocess;

template <typename AT>
struct service
{
	AT address;
	long unsigned program;
	long unsigned version;
	long unsigned port;
	protocol_t protocol;
	state_t state;

	service() = delete;
	service(const AT p_address, const long unsigned p_program, const long unsigned p_version, const long unsigned p_port, const protocol_t p_protocol, const state_t p_state);
	service(const service<AT> &r) = default;
	~service() = default;
	service &operator=(const service &) = default;

	bool operator<(const service<AT> &rs) const;
};

template <typename AT>
service<AT>::service(const AT p_address, const long unsigned p_program, const long unsigned p_version, const long unsigned p_port, const protocol_t p_protocol, const state_t p_state) : address(p_address), program(p_program), version(p_version), port(p_port), protocol(p_protocol), state(p_state) {}

template <typename AT>
service<AT>::service(const service<AT> &r) : address(r.address), program(r.program), version(r.version), port(r.port), protocol(r.protocol), state(r.state) {}

template <typename AT>
bool service<AT>::operator<(const service<AT> &rs) const
{
	if (address.compare(rs.address)<0) return true;
	else if (program<rs.program) return true;
	else if (version<rs.version) return true;
	else if (port<rs.port) return true;
	else if (protocol<rs.protocol) return true;
	else if (state<rs.state) return true;
	else return false;
}


struct service_string;
struct service_shm_string;

struct service_string : public service<std::string>
{
	service_string() = delete;
	service_string(const std::string p_address, const long unsigned p_program, const long unsigned p_version, const long unsigned p_port, const protocol_t p_protocol, const state_t p_state);
	service_string(const service_shm_string &rs);
	service_string(const service_string &) = default;
	~service_string() = default;
	service_string &operator=(const service_string &) = default;
};

struct service_shm_string : public service<shm_string>
{
	service_shm_string() = delete;
	service_shm_string(const shm_string p_address, const long unsigned p_program, const long unsigned p_version, const long unsigned p_port, const protocol_t p_protocol, const state_t p_state);
	service_shm_string(const service_string &rs, shm_char_allocator &al);
	service_shm_string(const service_shm_string &) = default;
	~service_shm_string() = default;
	service_shm_string &operator=(const service_shm_string &) = default;
};

typedef std::set<service_string> stdset;
typedef bi::managed_shared_memory::allocator<service_shm_string>::type shm_service_shm_string_allocator;
typedef bi::set<service_shm_string, std::less<service_shm_string>, shm_service_shm_string_allocator> biset;

void std_set_to_bi_set(bi::managed_shared_memory &segment, const stdset &si, biset &so);
void bi_set_to_std_set(bi::managed_shared_memory &segment, const biset &si, stdset &so);

struct monitor
{
	bi::offset_ptr<biset> m_services;

	monitor() = default;
	monitor(const monitor &) = delete;
	~monitor() = default;
	monitor &operator=(const monitor &) = delete;
	void init(bi::managed_shared_memory &m_segment);
	void deinit(bi::managed_shared_memory &m_segment);
};

std::ostream &operator<<(std::ostream &os, const service_string &r);

}

}

#endif
