/*
 * Copyright (c) 2009, 2010, Maciej Smolenski (jezdnia@gmail.com)
 * All rights reserved.
 *
 * This file is part of Oropo project.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 *     * Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *     * Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in the
 *     documentation and/or other materials provided with the distribution.
 *     * Neither the name of the Oropo nor the names of its contributors may
 *     be used to endorse or promote products derived from this software
 *     without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS
 * BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY,
 * OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF
 * THE POSSIBILITY OF SUCH DAMAGE.
 */
#ifndef _OROPO_RPC_CLIENT_HPP_
#define _OROPO_RPC_CLIENT_HPP_

#include <sys/types.h>
#include <sys/socket.h>
#include <netdb.h>
#include <sstream>
#include <cerrno>
#include <string>
#include <rpc/rpc.h>
#include <oropo/rpc/error.hpp>

#define PORTMAP_SERVICE_NAME "sunrpc"

namespace oropo
{

namespace rpc
{

template <typename CT>
class client
{
protected:
	std::string m_host;
	long unsigned m_program;
	long unsigned m_version;
	CLIENT *m_client;
	int m_sd;

	client() = delete;
	client(std::string p_host, long unsigned p_program, long unsigned p_version) throw(error);
	client(const client &) = delete;
	~client();
	client &operator=(const client &) = delete;
};

template <typename CT>
client<CT>::client(std::string p_host, long unsigned p_program, long unsigned p_version) throw(error) : m_host(p_host), m_program(p_program), m_version(p_version), m_client(NULL), m_sd(-1)
{
	struct addrinfo filter;
	struct addrinfo *addrs, *a_it;
	struct sockaddr_in pmaddr;
	int result;

	memset(&filter, 0, sizeof(filter));
	filter.ai_family = AF_INET;
	filter.ai_socktype = CT::socket_type;
	filter.ai_protocol = CT::protocol;

	if ((result = getaddrinfo(m_host.c_str(), PORTMAP_SERVICE_NAME, &filter, &addrs))!=0)
	{
		std::ostringstream os;
		os << "getaddrinfo failed host: " << m_host << " error: " << gai_strerror(result);
		throw error(make_error_code(RPC_FAILED), os.str());
	}

	for (a_it = addrs; a_it!=NULL; a_it = a_it->ai_next)
	{
		if ((m_sd = socket(a_it->ai_family, a_it->ai_socktype, a_it->ai_protocol)) == -1)
			continue;

		if (connect(m_sd, a_it->ai_addr, a_it->ai_addrlen) == 0)
			break;

		close(m_sd);
	}

	if (a_it==NULL)
	{
		freeaddrinfo(addrs);
		throw error(make_error_code(RPC_FAILED), std::string("can't connect to rpc_service on host: ") + m_host);
	}

	close(m_sd);

	memcpy(&pmaddr, a_it->ai_addr, a_it->ai_addrlen);
	freeaddrinfo(addrs);

	pmaddr.sin_port = 0;
	m_sd = RPC_ANYSOCK;
	if ((m_client = CT::create_client(&pmaddr, m_program, m_version, &m_sd)) == NULL)
	{
		std::ostringstream os;
		os << "create_client failed " << "host: " << m_host << " program: " << m_program << " version: " << m_version << " " << clnt_spcreateerror("");
		throw error(make_error_code(RPC_FAILED), os.str());
	}
}


template <typename CT>
client<CT>::~client()
{
	if (m_client) clnt_destroy(m_client);
}

}

}

#endif
