/*
 * Copyright (c) 2009, 2010, Maciej Smolenski (jezdnia@gmail.com)
 * All rights reserved.
 *
 * This file is part of Oropo project.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 *     * Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *     * Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in the
 *     documentation and/or other materials provided with the distribution.
 *     * Neither the name of the Oropo nor the names of its contributors may
 *     be used to endorse or promote products derived from this software
 *     without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS
 * BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY,
 * OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF
 * THE POSSIBILITY OF SUCH DAMAGE.
 */
#ifndef _OROPO_MANAGER_WORKER_SERVICE_HPP_
#define _OROPO_MANAGER_WORKER_SERVICE_HPP_

#include <boost/interprocess/sync/scoped_lock.hpp>
#include <boost/interprocess/sync/named_mutex.hpp>
#include <algorithm>
#include <functional>
#include <oropo/base/share.hpp>
#include <oropo/worker/data.hpp>

namespace oropo
{
namespace worker
{

namespace oro=oropo;
namespace bi=boost::interprocess;

template <typename CT>
class service : public oro::share<CT>
{
	constructor_initializer_helper<data *, CT> m_data;

public:
	service() = delete;
	service(const char *p_memory_name, const int p_memory_size, const char *p_mutex_name);
	service(const char *p_memory_name, const char *p_mutex_name);
	service(const service &) = delete;
	~service() = default;
	service &operator=(const service &) = delete;

	void get_informations(stdset &so);
	void set_informations(stdset &si);
	void add_information(mem_worker_data &p_wd);
	void remove_information(mem_worker_data &p_wd);
	template <typename PT>
	bool atomic_add_information_if_none(mem_worker_data const &p_wd, PT const &p_pred);
};

template <typename CT>
service<CT>::service(const char *p_memory_name, const int p_memory_size, const char *p_mutex_name)
		: oro::share<CT>(p_memory_name, p_memory_size, p_mutex_name), m_data(oro::share<CT>::m_memory.m_t)
{
}

template <typename CT>
service<CT>::service(const char *p_memory_name, const char *p_mutex_name)
		: oro::share<CT>(p_memory_name, p_mutex_name), m_data(oro::share<CT>::m_memory.m_t)
{
}

template <typename CT>
void service<CT>::get_informations(stdset &so)
{
	bi::managed_shared_memory &segment = share<CT>::m_memory.m_t;
	bi::scoped_lock<bi::named_mutex> sl(oro::share<CT>::m_mutex.m_t);
	bi_set_to_std_set(segment, *m_data.m_t->m_informations, so);
}

template <typename CT>
void service<CT>::set_informations(stdset &si)
{
	bi::managed_shared_memory &segment = share<CT>::m_memory.m_t;
	bi::scoped_lock<bi::named_mutex> sl(oro::share<CT>::m_mutex.m_t);
	std_set_to_bi_set(segment, si, *m_data.m_t->m_informations);
}

template <typename CT>
void service<CT>::add_information(mem_worker_data &p_wd)
{
	bi::managed_shared_memory &segment = share<CT>::m_memory.m_t;
	shm_worker_data swd(segment, p_wd);
	bi::scoped_lock<bi::named_mutex> sl(oro::share<CT>::m_mutex.m_t);
	m_data.m_t->m_informations->erase(swd);
	m_data.m_t->m_informations->insert(swd);
}

template <typename CT>
void service<CT>::remove_information(mem_worker_data &p_wd)
{
	bi::managed_shared_memory &segment = share<CT>::m_memory.m_t;
	shm_worker_data swd(segment, p_wd);
	bi::scoped_lock<bi::named_mutex> sl(oro::share<CT>::m_mutex.m_t);
	m_data.m_t->m_informations->erase(swd);
}

template <typename CT>
template <typename PT>
bool service<CT>::atomic_add_information_if_none(mem_worker_data const &p_wd, PT const &p_pred)
{
	bi::managed_shared_memory &segment = share<CT>::m_memory.m_t;
	bi::scoped_lock<bi::named_mutex> sl(oro::share<CT>::m_mutex.m_t);
	biset &infs = *m_data.m_t->m_informations;
	for (biset::iterator iter = infs.begin(); iter != infs.end(); iter++)
	{
		if (p_pred(*iter))
			return false;
	}
	shm_worker_data swd(segment, p_wd);
	m_data.m_t->m_informations->erase(swd);
	m_data.m_t->m_informations->insert(swd);
	return true;
}


}
}

#endif /* #ifndef _OROPO_MANAGER_WORKER_SERVICE_HPP_ */
