/*
 * Copyright (c) 2009, 2010, Maciej Smolenski (jezdnia@gmail.com)
 * All rights reserved.
 *
 * This file is part of Oropo project.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 *     * Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *     * Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in the
 *     documentation and/or other materials provided with the distribution.
 *     * Neither the name of the Oropo nor the names of its contributors may
 *     be used to endorse or promote products derived from this software
 *     without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS
 * BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY,
 * OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF
 * THE POSSIBILITY OF SUCH DAMAGE.
 */
#ifndef _OROPO_WORKER_WORKER_DATA_HPP_
#define _OROPO_WORKER_WORKER_DATA_HPP_

#include <ostream>
#include <boost/type_traits.hpp>
#include <boost/interprocess/offset_ptr.hpp>
#include <boost/interprocess/managed_shared_memory.hpp>
#include <oropo/monitor/monitor_data.hpp>
#include <oropo/manager/request.hpp>

namespace oropo
{
namespace worker
{

enum class state
{
	started, ready, waiting_for_request, waiting_for_service, checking_maybe_processing, skipping_request, returning_result, ended
};

std::ostream &operator<<(std::ostream &os, const state &p_state);

namespace oro=oropo;
namespace oromo=oro::monitor;
namespace oroma=oro::manager;
namespace bi=boost::interprocess;

template <typename ID>
class worker_data
{
	ID m_id;
	state m_state;
public:
	worker_data() = delete;
	worker_data(ID const p_id, state const p_state);
	worker_data(const worker_data &) = default;
	~worker_data() = default;
	worker_data &operator=(const worker_data &) = default;

	bool operator<(const worker_data<ID> &p_worker_data) const;

	ID get_id();
	state get_state();
	void set_id(ID const p_id);
	void set_state(state const p_state);

	template <typename T>
	friend std::ostream &operator<<(std::ostream &os, const worker_data<T> &mwd);
};

template <typename ID>
worker_data<ID>::worker_data(ID const p_id, state const p_state) : m_id(p_id), m_state(p_state) {}

template <typename ID>
bool worker_data<ID>::operator<(const worker_data<ID> &p_worker_data) const
{
	if (m_id<p_worker_data.m_id) return true;
	return false;
}

template <typename ID>
ID worker_data<ID>::get_id()
{
	return m_id;
}

template <typename ID>
state worker_data<ID>::get_state()
{
	return m_state;
}

template <typename ID>
void worker_data<ID>::set_id(ID const p_id)
{
	m_id = p_id;
}

template <typename ID>
void worker_data<ID>::set_state(state const p_state)
{
	m_state = p_state;
}

template <typename T>
std::ostream &operator<<(std::ostream &os, const worker_data<T> &mwd)
{
	os << " id: " << mwd.m_id;
	os << " state: " << mwd.m_state;
	return os;
}


class mem_worker_data;
class shm_worker_data;

class mem_worker_data : public worker_data<unsigned long>
{
	oromo::service_string *m_service;
	oroma::mem_request *m_request;

private:
	template <typename S, typename R>
	void initialize(S const * const p_service, R const * const p_request);
	void deinitialize();

public:
	mem_worker_data() = delete;
	mem_worker_data(unsigned long const p_id, state const p_state, oromo::service_string const * const p_service, oroma::mem_request const * const p_request);
	mem_worker_data(shm_worker_data const &swd);
	mem_worker_data(mem_worker_data const &swd);
	~mem_worker_data();
	mem_worker_data &operator=(const mem_worker_data &);

	oromo::service_string *get_service() const;
	oroma::mem_request *get_request() const;

	friend std::ostream &operator<<(std::ostream &os, const mem_worker_data &mwd);

	friend class shm_worker_data;
};

class shm_worker_data : public worker_data<unsigned long>
{
	bi::offset_ptr<oromo::service_shm_string> m_service;
	bi::offset_ptr<oroma::shm_request> m_request;
	bi::managed_shared_memory &m_segment;

private:
	void initialize(bi::managed_shared_memory &p_segment, oromo::service_string const * const p_service, oroma::mem_request const * const p_request);
	void initialize(bi::managed_shared_memory &p_segment, oromo::service_shm_string const * const p_service, oroma::shm_request const * const p_request);
	void deinitialize(bi::managed_shared_memory &p_segment);
	shm_worker_data &operator=(const shm_worker_data &) = delete;

public:
	shm_worker_data() = delete;
	shm_worker_data(bi::managed_shared_memory &p_segment, unsigned long const p_id, state const p_state, oromo::service_shm_string const * const p_service, oroma::shm_request const * const p_request);
	shm_worker_data(bi::managed_shared_memory &p_segment, mem_worker_data const &mwd);
	shm_worker_data(shm_worker_data const &mwd);
	~shm_worker_data();

	oromo::service_shm_string *get_service() const;
	oroma::shm_request *get_request() const;

	friend class mem_worker_data;
};


}
}


#endif /* #ifndef _OROPO_WORKER_WORKER_DATA_HPP_ */
